'' =================================================================================================
''
''   File....... jm_1-wire.spin
''   Purpose.... Essential 1-Wire interface routines
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2009-2025 Jon McPhalen
''               -- elements borrowed from Cam Thompson and Micah Dowty
''               -- see below for terms of use
''   E-mail.....
''   Started.... 29 JUL 2009
''   Updated.... 28 FEB 2025
''
'' =================================================================================================


con { fixed io pins }

  RX_PGM   = 31  { I }                                          ' serial / programming
  TX_PGM   = 30  { O }

  SDA1     = 29  { I/O }                                        ' i2c / eeprom
  SCL1     = 28  { I/O }


con

  #0, OW_DONE, OW_RESET, OW_WRITE, OW_READ, OW_RDBIT, OW_CRC    ' high-level 1W commands


con { reset codes }

  BUS_SHORT  = %00
  BAD_RESET  = %01
  GOOD_RESET = %10
  NO_DEVICE  = %11


con { 1W commands  }

  READ_ROM   = $33                                              ' 1W ROM commands
  MATCH_ROM  = $55
  SKIP_ROM   = $CC
  SRCH_ROM   = $F0
  ALARM_SRCH = $EC


var

  long  cog                                                     ' cog used by 1W driver

  long  owcmd                                                   ' command to 1-Wire interface
  long  owio                                                    ' data in/out
  long  owpin                                                   ' pin used
  long  us001                                                   ' ticks per microsecond


pub null

'' This is not a top-level object


pub start(pin)

'' Starts 1-Wire object

  stop                                                          ' stop if already running

  if ((pin => 0) and (pin =< 27))                               ' protect rx, tx, i2c
    owpin := pin                                                ' set pin to use
    us001 := clkfreq / 1_000_000                                ' set ticks per microsecond

    dira[pin] := 0                                              ' float OW pin
    owcmd := 0                                                  ' clear command
    cog := cognew(@onewire, @owcmd) + 1                         ' start cog

  return cog


pub stop

'' Stops 1-Wire driver; frees a cog

  if (cog)
    cogstop(cog - 1)
    cog := 0


pub reset

'' Resets 1-Wire bus; returns bus status
''
''   %00 = bus short
''   %01 = bad response; possible interference on bus
''   %10 = good bus & presence detection
''   %11 = no device

  owio := %11                                                   ' assume no device
  owcmd := OW_RESET                                             ' execute command
  repeat while owcmd                                            ' wait until finished

  return owio                                                   ' return bus status


pub write(b)

'' Write byte b to 1-Wire bus

  owio := b & $FF                                               ' cleanup byte
  owcmd := OW_WRITE                                             ' execute command
  repeat while owcmd                                            ' wait until finished


pub read

'' Reads byte from 1-Wire bus

  owio := 0                                                     ' clear old result
  owcmd := OW_READ                                              ' execute command
  repeat while owcmd                                            ' wait until finished

  return owio                                                   ' return 1-wire byte


pub rd_bit

'' Reads bit from 1-Wire bus
'' -- useful for monitoring device busy status

  owio := 0                                                     ' clear old result
  owcmd := OW_RDBIT                                             ' execute command
  repeat while owcmd                                            ' wait until finished

  return owio                                                   ' return 1-wire bit


pub crc8(p_src, n) : crc

'' Returns CRC8 of n bytes at pntr
'' -- PASM implementation by Cam Thompson

  owio := p_src                                                 ' store address of array
  owcmd := (n << 8) + OW_CRC                                    ' pack count into command
  repeat while owcmd                                            ' wait until finished

  return owio                                                   ' return crc


pub crc8x(p_src, len) : crc | b

'' Returns CRC8 of n bytes at p_src

  repeat len
    crc ^= byte[p_src++]
    repeat 8
      crc ->= 1
      if (crc & $8000_0000)
        crc ^= $8C
    crc &= $FF


dat

                        org       0

onewire                 mov       t1, par                       ' get start of structure
                        mov       cmdpntr, t1                   ' save hub addr of command
                        add       t1, #4
                        mov       iopntr, t1                    ' save hub addr of io value
                        add       t1, #4
                        rdlong    t2, t1                        ' read pin #
                        mov       owmask, #1                    ' create pin mask
                        shl       owmask, t2
                        add       t1, #4
                        rdlong    ustix, t1                     ' read ticks per microseconds

owmain                  rdlong    t1, cmdpntr           wz      ' get command
        if_z            jmp       #owmain                       ' wait for valid command
                        mov       bytecount, t1                 ' make copy (for crc byte count)
                        and       t1, #$FF                      ' strip off count

                        min       t1, #0                        ' limit range
                        max       t1, #6                        ' commands + 1
                        add       t1, #jmptable                 ' prep for jump
                        jmp       t1                            ' do it

jmptable                jmp       #cmdexit                      ' invalid command
do_cmd1                 jmp       #owreset                      ' reset 1-wire bus
do_cmd2                 jmp       #owwrbyte                     ' write a byte
do_cmd3                 jmp       #owrdbyte                     ' read a byte
do_cmd4                 jmp       #owrdbit                      ' read a bit
do_cmd5                 jmp       #owcalccrc                    ' calculate CRC

cmdexit                 wrlong    ZERO, cmdpntr                 ' clear old command
                        jmp       #owmain



' -----------------
' Reset 1-Wire bus
' -----------------
'
' %00 = bus short
' %01 = bad response; possible interference on bus
' %10 = good bus & presence detection
' %11 = no device

owreset                 mov       value, #%11                   ' assume no device
                        or        dira, owmask                  ' bus low
                        mov       usecs, #500                   ' reset pulse
                        call      #delayus
                        andn      dira, owmask                  ' release bus
                        mov       usecs, #5
                        call      #delayus
                        test      owmask, ina           wc      ' sample for short, 1W -> C
                        muxc      value, #%10                   ' C -> value.1
                        mov       usecs, #70
                        call      #delayus
                        test      owmask, ina           wc      ' sample for presence, 1W -> C
                        muxc      value, #%01                   ' C -> value.0
                        mov       usecs, #400
                        call      #delayus
                        wrlong    value, iopntr                 ' update hub
                        jmp       #cmdexit



' -------------------------
' Write byte to 1-Wire bus
' -------------------------
'
owwrbyte                rdlong    value, iopntr                 ' get byte from hub
                        mov       bitcount, #8                  ' write 8 bits
wrloop                  shr       value, #1             wc      ' value.0 -> C, value >>= 1
        if_c            mov       usecs, #6                     ' write 1
        if_nc           mov       usecs, #60                    ' write 0
                        or        dira, owmask                  ' pull bus low
                        call      #delayus
                        andn      dira, owmask                  ' release bus
        if_c            mov       usecs, #64                    ' pad for 1
        if_nc           mov       usecs, #10                    ' pad for 0
                        call      #delayus
                        djnz      bitcount, #wrloop             ' all bits done?
                        jmp       #cmdexit



' --------------------------
' Read byte from 1-Wire bus
' --------------------------
'
owrdbyte                mov       bitcount, #8                  ' read 8 bits
                        mov       value, #0                     ' clear workspace
rdloop                  or        dira, owmask                  ' bus low
                        mov       usecs, #6
                        call      #delayus
                        andn      dira, owmask                  ' release bus
                        mov       usecs, #9                     ' hold-off before sample
                        call      #delayus
                        test      owmask, ina           wc      ' sample bus, 1W -> C
                        shr       value, #1                     ' value >>= 1
                        muxc      value, #%1000_0000            ' C -> value.7
                        mov       usecs, #55                    ' finish read slot
                        call      #delayus
                        djnz      bitcount, #rdloop             ' all bits done?
                        wrlong    value, iopntr                 ' yes, update hub
                        jmp       #cmdexit



' -------------------------
' Read bit from 1-Wire bus
' -------------------------
'
owrdbit                 mov       value, #0                     ' clear workspace
                        or        dira, owmask                  ' bus low
                        mov       usecs, #6
                        call      #delayus
                        andn      dira, owmask                  ' release bus
                        mov       usecs, #9                     ' hold-off before sample
                        call      #delayus
                        test      owmask, ina           wc      ' sample bus, 1W -> C
                        muxc      value, #1                     ' C -> value.0
                        mov       usecs, #55                    ' finish read slot
                        call      #delayus
                        wrlong    value, iopntr                 ' update hub
                        jmp       #cmdexit



' ----------------------
' Calculate CRC8
' * code by Cam Thompson
' ----------------------
'
owcalccrc               mov       value, #0                     ' clear workspace
                        shr       bytecount, #8         wz      ' clear command from count
        if_z            jmp       #savecrc
                        rdlong    hubpntr, iopntr               ' get address of array

crcbyte                 rdbyte    t1, hubpntr                   ' read byte from array
                        add       hubpntr, #1                   ' point to next
                        mov       bitcount, #8

crcbit                  mov       t2, t1                        ' x^8 + x^5 + x^4 + 1
                        shr       t1, #1
                        xor       t2, value
                        shr       value, #1
                        shr       t2, #1                wc
        if_c            xor       value, #$8C
                        djnz      bitcount, #crcbit
                        djnz      bytecount, #crcbyte

savecrc                 wrlong    value, iopntr                 ' update hub
                        jmp       #cmdexit



' ---------------------
' Delay in microseconds
' ---------------------
'
delayus                 mov       ustimer, cnt                  ' sync with system clock
                        sub       ustimer, #24                  ' account for setup/call/return
                        add       ustimer, ustix                ' set timer for 1us
usloop                  waitcnt   ustimer, ustix                ' wait and reload
                        djnz      usecs, #usloop                ' update delay count
delayus_ret             ret


' --------------------------------------------------------------------------------------------------

ZERO                    long      0

cmdpntr                 res       1                             ' hub address of command
iopntr                  res       1                             ' hub address of io value
owmask                  res       1                             ' mask of pin used
ustix                   res       1                             ' clock ticks in 1us

value                   res       1
bitcount                res       1
bytecount               res       1
hubpntr                 res       1
usecs                   res       1
ustimer                 res       1

t1                      res       1
t2                      res       1

                        fit       496


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}